@@ -1,7 +1,7 @@
 /*
   Copyright (C) 1994, 1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002,
-  2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012 Free
-  Software Foundation, Inc.
+  2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013,
+  2014, 2015 Free Software Foundation, Inc.
 
   This file is part of GNU Inetutils.
 
@@ -98,6 +98,7 @@
 #include <glob.h>
 #include <argp.h>
 #include <error.h>
+#include <xgetcwd.h>
 
 #include <progname.h>
 #include <libinetutils.h>
@@ -131,9 +132,11 @@ socklen_t data_dest_len;
 struct sockaddr_storage his_addr;	/* Peer address.  */
 socklen_t his_addrlen;
 int logging;			/* Enable log to syslog.  */
+int no_version;			/* Don't print version to client.  */
 int type = TYPE_A;		/* Default TYPE_A.  */
 int form = FORM_N;		/* Default FORM_N.  */
 int debug;			/* Enable debug mode if 1.  */
+int rfc2577 = 1;		/* Follow suggestions in RFC 2577.  */
 int timeout = 900;		/* Timeout after 15 minutes of inactivity.  */
 int maxtimeout = 7200;		/* Don't allow idle time to be set
 				   beyond 2 hours.  */
@@ -159,7 +162,6 @@ static jmp_buf urgcatch;
 static int stru = STRU_F;	/* Avoid C keyword.  */
 static int stru_mode = MODE_S;	/* Default STRU mode stru_mode = MODE_S.  */
 static int anon_only;		/* Allow only anonymous login.  */
-static int no_version;		/* Don't print version to client.  */
 static int daemon_mode;		/* Start in daemon mode.  */
 static off_t file_size;
 static off_t byte_count;
@@ -195,7 +197,7 @@ off_to_str (off_t off)
   if (sizeof (off) > sizeof (long))
     sprintf (*next_buf, "%lld", (long long int) off);
   else if (sizeof (off) == sizeof (long))
-    sprintf (*next_buf, "%ld", off);
+    sprintf (*next_buf, "%ld", (long) off);
   else
     sprintf (*next_buf, "%d", (int) off);
 
@@ -252,11 +254,20 @@ static FILE *getdatasock (const char *);
 static char *gunique (const char *);
 static void lostconn (int);
 static void myoob (int);
-static int receive_data (FILE *, FILE *);
+static int receive_data (FILE *, FILE *, off_t);
 static void send_data (FILE *, FILE *, off_t);
 static void sigquit (int);
 
-const char doc[] = "File Transfer Protocol Daemon";
+const char doc[] =
+#ifdef WITH_PAM
+  "File Transfer Protocol daemon, offering PAM service 'ftp'.";
+#else
+  "File Transfer Protocol daemon.";
+#endif
+
+enum {
+  OPT_NONRFC2577 = CHAR_MAX + 1,
+};
 
 static struct argp_option options[] = {
 #define GRID 0
@@ -269,6 +280,12 @@ static struct argp_option options[] = {
   { "debug", 'd', NULL, 0,
     "debug mode",
     GRID+1 },
+  { "ipv4", '4', NULL, 0,
+    "restrict daemon to IPv4",
+    GRID+1 },
+  { "ipv6", '6', NULL, 0,
+    "restrict daemon to IPv6",
+    GRID+1 },
   { "logging", 'l', NULL, 0,
     "increase verbosity of syslog messages",
     GRID+1 },
@@ -284,10 +301,13 @@ static struct argp_option options[] = {
   { "max-timeout", 'T', NULL, 0,
     "reset maximum value of timeout allowed",
     GRID+1 },
+  { "non-rfc2577", OPT_NONRFC2577, NULL, 0,
+    "neglect RFC 2577 by giving info on missing users",
+    GRID+1 },
   { "umask", 'u', "VAL", 0,
     "set default umask",
     GRID+1 },
-  { "auth", 'a', "AUTH", OPTION_ARG_OPTIONAL,
+  { "auth", 'a', "AUTH", 0,
     "use AUTH for authentication",
     GRID+1 },
   { NULL, 0, NULL, 0, "AUTH can be one of the following:", GRID+2 },
@@ -296,7 +316,7 @@ static struct argp_option options[] = {
     GRID+3 },
 #ifdef WITH_PAM
   { "  pam", 0, NULL, OPTION_DOC|OPTION_NO_TRANS,
-    "using pam 'ftp' module",
+    "using PAM service 'ftp'",
     GRID+3 },
 #endif
 #ifdef WITH_KERBEROS
@@ -305,7 +325,7 @@ static struct argp_option options[] = {
     GRID+3 },
 #endif
 #ifdef WITH_KERBEROS5
-  { "  kderberos5", 0, NULL, OPTION_DOC|OPTION_NO_TRANS,
+  { "  kerberos5", 0, NULL, OPTION_DOC|OPTION_NO_TRANS,
     "",
     GRID+3 },
 #endif
@@ -314,7 +334,7 @@ static struct argp_option options[] = {
     "",
     GRID+3 },
 #endif
-  { NULL }
+  { NULL, 0, NULL, 0, NULL, 0 }
 };
 
 static error_t
@@ -322,6 +342,16 @@ parse_opt (int key, char *arg, struct argp_state *state)
 {
   switch (key)
     {
+    case '4':
+      /* Active in daemon mode only.  */
+      usefamily = AF_INET;
+      break;
+
+    case '6':
+      /* Active in daemon mode only.  */
+      usefamily = AF_INET6;
+      break;
+
     case 'A':
       /* Anonymous ftp only.  */
       anon_only = 1;
@@ -398,6 +428,10 @@ parse_opt (int key, char *arg, struct argp_state *state)
 	break;
       }
 
+    case OPT_NONRFC2577:
+      rfc2577 = 0;
+      break;
+
     default:
       return ARGP_ERR_UNKNOWN;
     }
@@ -439,7 +473,9 @@ main (int argc, char *argv[], char **envp)
 #ifdef HAVE_INITSETPROCTITLE
   /* Save start and extent of argv for setproctitle.  */
   initsetproctitle (argc, argv, envp);
-#endif /* HAVE_INITSETPROCTITLE */
+#else /* !HAVE_INITSETPROCTITLE */
+  (void) envp;		/* Silence warnings.  */
+#endif
 
   /* Parse the command line */
   iu_argp_init ("ftpd", default_program_authors);
@@ -552,14 +588,14 @@ main (int argc, char *argv[], char **envp)
       exit (EXIT_SUCCESS);
     }
 
-  /* Display a Welcome message if exists,
-     N.B. reply(220,) must follow.  */
-  display_file (PATH_FTPWELCOME, 220);
-
   hostname = localhost ();
   if (!hostname)
     perror_reply (550, "Local resource failure: malloc");
 
+  /* Display a Welcome message if it exists.
+     N.B. a reply(220,) must follow as continuation.  */
+  display_file (PATH_FTPWELCOME, 220);
+
   /* Tell them we're ready to roll.  */
   if (!no_version)
     reply (220, "%s FTP server (%s %s) ready.",
@@ -580,7 +616,7 @@ static char *
 curdir (void)
 {
   static char *path = 0;
-  extern char *xgetcwd (void);
+
   free (path);
   path = xgetcwd ();
   if (!path)
@@ -641,10 +677,12 @@ sgetsave (const char *s)
 static void
 complete_login (struct credentials *pcred)
 {
+  char *cwd;
+
   if (setegid ((gid_t) pcred->gid) < 0)
     {
       reply (550, "Can't set gid.");
-      return;
+      goto bad;
     }
 
 #ifdef HAVE_INITGROUPS
@@ -652,7 +690,7 @@ complete_login (struct credentials *pcred)
 #endif
 
   /* open wtmp before chroot */
-  snprintf (ttyline, sizeof (ttyline), "ftp%d", getpid ());
+  snprintf (ttyline, sizeof (ttyline), "ftp%d", (int) getpid ());
   logwtmp_keep_open (ttyline, pcred->name, pcred->remotehost);
 
   if (pcred->guest)
@@ -660,6 +698,9 @@ complete_login (struct credentials *pcred)
       /* We MUST do a chdir () after the chroot. Otherwise
          the old current directory will be accessible as "."
          outside the new root!  */
+      char *subdir = strstr(pcred->rootdir, "/./");
+      if (subdir)
+      	*subdir++ = '\0';
       if (chroot (pcred->rootdir) < 0)
 	{
 	  reply (550, "Can't set guest privileges.");
@@ -675,9 +716,17 @@ complete_login (struct credentials *pcred)
 	  reply(550, "Can't access home directory.");
 	  goto bad;
 	}
+      if (subdir && chdir (subdir) < 0)
+	{
+	  reply(550, "Can't change directory to %s.", subdir);
+	  goto bad;
+	}
     }
   else if (pcred->dochroot)
     {
+      char *subdir = strstr(pcred->rootdir, "/./");
+      if (subdir)
+      	*subdir++ = '\0';
       if (chroot (pcred->rootdir) < 0)
 	{
 	  reply (550, "Can't change root.");
@@ -693,30 +742,48 @@ complete_login (struct credentials *pcred)
 	  reply(550, "Can't access home directory.");
 	  goto bad;
 	}
-      setenv ("HOME", pcred->homedir, 1);
-    }
-  else
-    {
-      if (seteuid ((uid_t)pcred->uid) < 0)
+      if (subdir && chdir (subdir) < 0)
 	{
-	  reply(550, "Can't set uid.");
+	  reply(550, "Can't change directory to %s.", subdir);
 	  goto bad;
 	}
+    }
+
+  if (seteuid ((uid_t) pcred->uid) < 0)
+    {
+      reply (550, "Can't set uid.");
+      goto bad;
+    }
+
+  if (!pcred->guest && !pcred->dochroot)	/* Remaining case.  */
+    {
+      if (seteuid ((uid_t)pcred->uid) < 0)
+ 	{
+	  reply(550, "Can't set uid.");
+ 	  goto bad;
+ 	}
       if (chdir (pcred->rootdir) < 0)
 	{
-          if (chdir ("/") < 0)
+	  if (chdir ("/") < 0)
 	    {
 	      reply (530, "User %s: can't change directory to %s.",
-		 pcred->name, pcred->homedir);
+		     pcred->name, pcred->homedir);
 	      goto bad;
 	    }
-          else
-	    lreply (230, "No directory! Logging in with home=/");
+
+	  lreply (230, "No directory! Logging in with home=/");
 	}
+    }
+
+  cwd = xgetcwd ();
+  if (cwd)
+    {
+      setenv ("HOME", cwd, 1);
+      free (cwd);
     }
 
   /* Display a login message, if it exists.
-     N.B. reply(230,) must follow the message.  */
+     N.B. a reply(230,) must follow after this message.  */
   display_file (PATH_FTPLOGINMESG, 230);
 
   if (pcred->guest)
@@ -746,7 +813,7 @@ complete_login (struct credentials *pcred)
   return;
 bad:
   /* Forget all about it... */
-  end_login (pcred);
+  end_login (pcred);	/* Resets pcred->logged_in.  */
 }
 
 /* USER command.
@@ -756,6 +823,8 @@ bad:
 void
 user (const char *name)
 {
+  int ret;
+
   if (cred.logged_in)
     {
       if (cred.guest || cred.dochroot)
@@ -767,7 +836,8 @@ user (const char *name)
     }
 
   /* Non zero means failed.  */
-  if (auth_user (name, &cred) != 0)
+  ret = auth_user (name, &cred);
+  if (!rfc2577 && ret != 0)
     {
       /* If they gave us a reason.  */
       if (cred.message)
@@ -783,6 +853,17 @@ user (const char *name)
 		cred.remotehost, name);
       return;
     }
+  else if (rfc2577 && ret != 0)
+    cred.delayed_reject = 1;
+  else
+    cred.delayed_reject = 0;
+
+  /* Only messages for anonymous guests are accepted.  */
+  if (rfc2577 && !cred.guest && cred.message)
+    {
+      free (cred.message);
+      cred.message = NULL;
+    }
 
   /* If the server is set to serve anonymous service only
      the request have to come from a guest or a chrooted.  */
@@ -800,7 +881,20 @@ user (const char *name)
 
   if (cred.message)
     {
-      reply (331, "%s", cred.message);
+      /* Stacked PAM modules for authentication may have
+       * produced a multiline message at this point.
+       * The FTP protocol does not cope well with this,
+       * so we transfer only the very last line, which
+       * should reflect the active authentication mechanism.
+       */
+      char *msg = strrchr (cred.message, '\n');
+
+      if (msg)
+	msg++;		/* Step over separator.  */
+      else
+	msg = cred.message;
+
+      reply (331, "%s", msg);
       free (cred.message);
       cred.message = NULL;
     }
@@ -822,9 +916,15 @@ end_login (struct credentials *pcred)
 {
   char *remotehost = pcred->remotehost;
   int atype = pcred->auth_type;
+
   seteuid (getuid ());
   if (pcred->logged_in)
-    logwtmp_keep_open (ttyline, "", "");
+    {
+      logwtmp_keep_open (ttyline, "", "");
+#ifdef WITH_PAM
+      pam_end_login (pcred);
+#endif
+    }
 
   free (pcred->name);
   if (pcred->passwd)
@@ -835,7 +935,7 @@ end_login (struct credentials *pcred)
   free (pcred->homedir);
   free (pcred->rootdir);
   free (pcred->shell);
-  if (pcred->pass)		/* ??? */
+  if (pcred->pass)		/* Properly erase old password.  */
     {
       memset (pcred->pass, 0, strlen (pcred->pass));
       free (pcred->pass);
@@ -844,6 +944,8 @@ end_login (struct credentials *pcred)
   memset (pcred, 0, sizeof (*pcred));
   pcred->remotehost = remotehost;
   pcred->auth_type = atype;
+  pcred->logged_in = 0;
+  pcred->delayed_reject = 0;
 }
 
 void
@@ -859,17 +961,31 @@ pass (const char *passwd)
   if (!cred.guest)		/* "ftp" is the only account allowed no password.  */
     {
       /* Try to authenticate the user.  Failed if != 0.  */
-      if (auth_pass (passwd, &cred) != 0)
+      if (auth_pass (passwd, &cred) != 0 || cred.delayed_reject)
 	{
-	  /* Any particular reasons.  */
-	  if (cred.message)
+	  /* Any particular reason?  */
+	  if (rfc2577)
+	    {
+	      if (cred.message)
+		{
+		  free (cred.message);
+		  cred.message = NULL;
+		}
+	      reply (530, "Login incorrect.");
+	    }
+	  else if (cred.message)
 	    {
 	      reply (530, "%s", cred.message);
 	      free (cred.message);
 	      cred.message = NULL;
 	    }
+	  else if (cred.expired & AUTH_EXPIRED_ACCT)
+	    reply (530, "Account is expired.");
+	  else if (cred.expired & AUTH_EXPIRED_PASS)
+	    reply (530, "Password has expired.");
 	  else
 	    reply (530, "Login incorrect.");
+
 	  if (logging)
 	    syslog (LOG_NOTICE, "FTP LOGIN FAILED FROM %s, %s",
 		    cred.remotehost, curname);
@@ -877,14 +993,28 @@ pass (const char *passwd)
 	    {
 	      syslog (LOG_NOTICE, "repeated login failures from %s",
 		      cred.remotehost);
+	      reply (421,
+		     "Service not available, closing control connection.");
 	      exit (EXIT_SUCCESS);
 	    }
 	  return;
 	}
+      if (cred.message)
+	{
+	  /* At least PAM might have committed additional messages.
+	   * Reply code 230 is used, since at this point the client
+	   * has been accepted.  */
+	  lreply_multiline (230, cred.message);
+	  free (cred.message);
+	  cred.message = NULL;
+	}
     }
   cred.logged_in = 1;		/* Everything seems to be allright.  */
   complete_login (&cred);
-  login_attempts = 0;		/* This time successful.  */
+  if (cred.logged_in)
+    login_attempts = 0;		/* This time successful.  */
+  else
+    ++login_attempts;
 }
 
 void
@@ -893,7 +1023,7 @@ retrieve (const char *cmd, const char *name)
   FILE *fin, *dout;
   struct stat st;
   int (*closefunc) (FILE *);
-  size_t buffer_size = 0;
+  size_t buffer_size = BUFSIZ;	/* Dynamic buffer.  */
 
   if (cmd == 0)
     {
@@ -901,9 +1031,8 @@ retrieve (const char *cmd, const char *name)
       if (type != TYPE_A)
 	fin = fopen (name, "rb"), closefunc = fclose;
       else
-#else
+#endif
       fin = fopen (name, "r"), closefunc = fclose;
-#endif /* __CYGWIN__ */
       st.st_size = 0;
     }
   else
@@ -914,7 +1043,6 @@ retrieve (const char *cmd, const char *name)
       name = line;
       fin = ftpd_popen (line, "r"), closefunc = ftpd_pclose;
       st.st_size = -1;
-      buffer_size = BUFSIZ;
     }
 
   if (fin == NULL)
@@ -935,6 +1063,9 @@ retrieve (const char *cmd, const char *name)
       reply (550, "%s: not a plain file.", name);
       goto done;
     }
+  else if (cmd == 0)
+    buffer_size = st.st_blksize;	/* Depends on file system.  */
+
   if (restart_point)
     {
       if (type == TYPE_A)
@@ -949,7 +1080,10 @@ retrieve (const char *cmd, const char *name)
 	      c = getc (fin);
 	      if (c == EOF)
 		{
-		  perror_reply (550, name);
+		  /* Error code 554 was introduced in RFC 1123.  */
+		  reply (554,
+			 "Action not taken: invalid REST value %jd for %s.",
+			 restart_point, name);
 		  goto done;
 		}
 	      if (c == '\n')
@@ -958,7 +1092,11 @@ retrieve (const char *cmd, const char *name)
 	}
       else if (lseek (fileno (fin), restart_point, SEEK_SET) < 0)
 	{
-	  perror_reply (550, name);
+	  if (errno == EINVAL)
+	    reply (554, "Action not taken: invalid REST value %jd for %s.",
+		   restart_point, name);
+	  else
+	    perror_reply (550, name);
 	  goto done;
 	}
     }
@@ -1014,7 +1152,7 @@ store (const char *name, const char *mode, int unique)
     }
   fout = fopen (name, nmode);
   closefunc = fclose;
-  if (fout == NULL)
+  if (fout == NULL || fstat (fileno (fout), &st) < 0)
     {
       perror_reply (553, name);
       LOGCMD (*mode == 'w' ? "put" : "append", name);
@@ -1035,7 +1173,10 @@ store (const char *name, const char *mode, int unique)
 	      c = getc (fout);
 	      if (c == EOF)
 		{
-		  perror_reply (550, name);
+		  /* Error code 554 was introduced in RFC 1123.  */
+		  reply (554,
+			 "Action not taken: invalid REST value %jd for %s.",
+			 restart_point, name);
 		  goto done;
 		}
 	      if (c == '\n')
@@ -1052,14 +1193,18 @@ store (const char *name, const char *mode, int unique)
 	}
       else if (lseek (fileno (fout), restart_point, SEEK_SET) < 0)
 	{
-	  perror_reply (550, name);
+	  if (errno == EINVAL)
+	    reply (554, "Action not taken: invalid REST value %jd for %s.",
+		   restart_point, name);
+	  else
+	    perror_reply (550, name);
 	  goto done;
 	}
     }
   din = dataconn (name, (off_t) - 1, "r");
   if (din == NULL)
     goto done;
-  if (receive_data (din, fout) == 0)
+  if (receive_data (din, fout, st.st_blksize) == 0)
     {
       if (unique)
 	reply (226, "Transfer complete (unique file name:%s).", name);
@@ -1119,7 +1264,8 @@ getdatasock (const char *mode)
 	goto bad;
       sleep (tries);
     }
-  seteuid ((uid_t) cred.uid);
+  if (seteuid ((uid_t) cred.uid) != 0)
+    _exit (EXIT_FAILURE);
 
 #if defined IP_TOS && defined IPTOS_THROUGHPUT && defined IPPROTO_IP
   if (ctrl_addr.ss_family == AF_INET)
@@ -1134,7 +1280,8 @@ getdatasock (const char *mode)
 bad:
   /* Return the real value of errno (close may change it) */
   t = errno;
-  seteuid ((uid_t) cred.uid);
+  if (seteuid ((uid_t) cred.uid) != 0)
+    _exit (EXIT_FAILURE);
   close (s);
   errno = t;
   return NULL;
@@ -1236,6 +1383,8 @@ dataconn (const char *name, off_t size, const char *mode)
   return file;
 }
 
+#define IU_MMAP_SIZE 0x800000	/* 8 MByte */
+
 /* Tranfer the contents of "instr" to "outstr" peer using the appropriate
    encapsulation of the data subject * to Mode, Structure, and Type.
 
@@ -1244,9 +1393,12 @@ static void
 send_data (FILE * instr, FILE * outstr, off_t blksize)
 {
   int c, cnt, filefd, netfd;
-  char *buf, *bp;
+  char *buf = MAP_FAILED;
   off_t curpos;
-  size_t len, filesize;
+#ifdef HAVE_MMAP
+  char *bp;
+  off_t len, filesize;
+#endif
 
   transflag++;
   if (setjmp (urgcatch))
@@ -1258,9 +1410,16 @@ send_data (FILE * instr, FILE * outstr, off_t blksize)
   netfd = fileno (outstr);
   filefd = fileno (instr);
 #ifdef HAVE_MMAP
-  if (file_size > 0)
+  /* Last argument in mmap() must be page aligned,
+   * at least for Solaris and Linux, so use mmap()
+   * only with null offset retrievals.
+   */
+  if (file_size > 0 && file_size < IU_MMAP_SIZE && restart_point == 0)
     {
       curpos = lseek (filefd, 0, SEEK_CUR);
+      if (debug)
+	syslog (LOG_DEBUG, "Position is %jd. Attempting mmap call.",
+		curpos);
       if (curpos >= 0)
 	{
 	  filesize = file_size - curpos;
@@ -1276,6 +1435,8 @@ send_data (FILE * instr, FILE * outstr, off_t blksize)
 #ifdef HAVE_MMAP
       if (file_size > 0 && curpos >= 0 && buf != MAP_FAILED)
 	{
+	  if (debug)
+	    syslog (LOG_DEBUG, "Reading file as ascii in mmap mode.");
 	  len = 0;
 	  while (len < filesize)
 	    {
@@ -1298,6 +1459,8 @@ send_data (FILE * instr, FILE * outstr, off_t blksize)
 	  return;
 	}
 #endif
+      if (debug)
+	syslog (LOG_DEBUG, "Reading file as ascii in byte mode.");
       while ((c = getc (instr)) != EOF)
 	{
 	  byte_count++;
@@ -1323,6 +1486,8 @@ send_data (FILE * instr, FILE * outstr, off_t blksize)
 #ifdef HAVE_MMAP
       if (file_size > 0 && curpos >= 0 && buf != MAP_FAILED)
 	{
+	  if (debug)
+	    syslog (LOG_DEBUG, "Reading file as image in mmap mode.");
 	  bp = buf;
 	  len = filesize;
 	  do
@@ -1349,6 +1514,16 @@ send_data (FILE * instr, FILE * outstr, off_t blksize)
 	  return;
 	}
 #endif
+      if (debug)
+	{
+	  syslog (LOG_DEBUG, "Reading file as image in block mode.");
+	  curpos = lseek (filefd, 0, SEEK_CUR);
+	  if (curpos < 0)
+	    syslog (LOG_DEBUG, "Input file: %m");
+	  else
+	    syslog (LOG_DEBUG, "Starting at position %jd.", curpos);
+	}
+
       if (blksize <= 0)
 	blksize = LARGE_TRANSFER_BLOCKSIZE;
 
@@ -1362,6 +1537,7 @@ send_data (FILE * instr, FILE * outstr, off_t blksize)
       while ((cnt = read (filefd, buf, (u_int) blksize)) > 0 &&
 	     write (netfd, buf, cnt) == cnt)
 	byte_count += cnt;
+
       transflag = 0;
       free (buf);
       if (cnt != 0)
@@ -1393,11 +1569,11 @@ file_err:
 
    N.B.: Form isn't handled.  */
 static int
-receive_data (FILE * instr, FILE * outstr)
+receive_data (FILE * instr, FILE * outstr, off_t blksize)
 {
   int c;
   int cnt, bare_lfs = 0;
-  char buf[BUFSIZ];
+  char *buf;
 
   transflag++;
   if (setjmp (urgcatch))
@@ -1409,12 +1585,24 @@ receive_data (FILE * instr, FILE * outstr)
     {
     case TYPE_I:
     case TYPE_L:
-      while ((cnt = read (fileno (instr), buf, sizeof (buf))) > 0)
+      buf = malloc ((u_int) blksize);
+      if (buf == NULL)
+	{
+	  transflag = 0;
+	  perror_reply (451, "Local resource failure: malloc");
+	  return -1;
+	}
+
+      while ((cnt = read (fileno (instr), buf, blksize)) > 0)
 	{
 	  if (write (fileno (outstr), buf, cnt) != cnt)
-	    goto file_err;
+	    {
+	      free (buf);
+	      goto file_err;
+	    }
 	  byte_count += cnt;
 	}
+      free (buf);
       if (cnt < 0)
 	goto data_err;
       transflag = 0;
@@ -1525,6 +1713,7 @@ statcmd (void)
   if (!isdigit (cred.remotehost[0]))
     printf (" (%s)", addrstr);
   printf ("\r\n");
+  printf ("     Session timeout is %d seconds\r\n", timeout);
   if (cred.logged_in)
     {
       if (cred.guest)
@@ -1591,12 +1780,15 @@ reply (int n, const char *fmt, ...)
   va_start (ap, fmt);
   printf ("%d ", n);
   vprintf (fmt, ap);
+  va_end (ap);
   printf ("\r\n");
   fflush (stdout);
   if (debug)
     {
       syslog (LOG_DEBUG, "<--- %d ", n);
+      va_start (ap, fmt);
       vsyslog (LOG_DEBUG, fmt, ap);
+      va_end (ap);
     }
 }
 
@@ -1607,12 +1799,51 @@ lreply (int n, const char *fmt, ...)
   va_start (ap, fmt);
   printf ("%d- ", n);
   vprintf (fmt, ap);
+  va_end (ap);
   printf ("\r\n");
   fflush (stdout);
   if (debug)
     {
       syslog (LOG_DEBUG, "<--- %d- ", n);
+      va_start (ap, fmt);
       vsyslog (LOG_DEBUG, fmt, ap);
+      va_end (ap);
+    }
+}
+
+/* Send a possibly multiline reply as individual
+ * lines of message with identical status code.
+ * No format string input!
+ */
+void
+lreply_multiline (int n, const char *text)
+{
+  char *line;
+
+  line = strdup (text);
+  if (line == NULL)
+    return;
+  else
+    {
+      int stop = 0;
+      char *p1 = line, *p2;
+
+      do
+	{
+	  p2 = strchrnul (p1, '\n');
+	  stop = (*p2 == '\0');		/* End of input string?  */
+	  *p2 = '\0';
+	  printf ("%d- ", n);
+	  printf ("%s\r\n", p1);
+	  if (debug)
+	    {
+	      syslog (LOG_DEBUG, "<--- %d- ", n);
+	      syslog (LOG_DEBUG, "%s", p1);
+	    }
+	  p1 = ++p2;			/* P1 is used within bounds.  */
+	}
+      while (!stop);
+      free (line);
     }
 }
 
@@ -1669,8 +1900,6 @@ cwd (const char *path)
 void
 makedir (const char *name)
 {
-  extern char *xgetcwd (void);
-
   LOGCMD ("mkdir", name);
   if (mkdir (name, 0777) < 0)
     perror_reply (550, name);
@@ -1706,8 +1935,8 @@ removedir (const char *name)
 void
 pwd (void)
 {
-  extern char *xgetcwd (void);
   char *path = xgetcwd ();
+
   if (path)
     {
       reply (257, "\"%s\" is current directory.", path);
@@ -1764,17 +1993,13 @@ dolog (struct sockaddr *sa, socklen_t salen, struct credentials *pcred)
 void
 dologout (int status)
 {
-  /* Racing condition with SIGURG: If SIGURG is receive
-     here, it will jump back has root in the main loop
+  /* Race condition with SIGURG: If SIGURG is received
+     here, it will jump back has root in the main loop.
      David Greenman:dg@root.com.  */
   transflag = 0;
+  end_login (&cred);
 
-  if (cred.logged_in)
-    {
-      seteuid (getuid ());
-      logwtmp_keep_open (ttyline, "", "");
-    }
-  /* beware of flushing buffers after a SIGPIPE */
+  /* Beware of flushing buffers after a SIGPIPE.  */
   _exit (status);
 }
 
@@ -1815,11 +2040,18 @@ myoob (int signo _GL_UNUSED_PARAMETER)
    a legitimate response by Jon Postel in a telephone conversation
    with Rick Adams on 25 Jan 89.  */
 void
-passive (void)
+passive (int epsv, int af)
 {
   char *p, *a;
+  int try_af;
 
-  pdata = socket (ctrl_addr.ss_family, SOCK_STREAM, 0);
+  /* EPSV might ask for a particular address family.  */
+  if (epsv == PASSIVE_EPSV && af > 0)
+    try_af = af;
+  else
+    try_af = ctrl_addr.ss_family;
+
+  pdata = socket (try_af, SOCK_STREAM, 0);
   if (pdata < 0)
     {
       perror_reply (425, "Can't open passive connection");
@@ -1827,6 +2059,7 @@ passive (void)
     }
   memcpy (&pasv_addr, &ctrl_addr, sizeof (pasv_addr));
   pasv_addrlen = ctrl_addrlen;
+
   /* Erase the port number.  */
   if (pasv_addr.ss_family == AF_INET6)
     ((struct sockaddr_in6 *) &pasv_addr)->sin6_port = 0;
@@ -1836,31 +2069,76 @@ passive (void)
   seteuid (getuid ());
   if (bind (pdata, (struct sockaddr *) &pasv_addr, pasv_addrlen) < 0)
     {
-      seteuid ((uid_t) cred.uid);
+      if (seteuid ((uid_t) cred.uid))
+	_exit (EXIT_FAILURE);
       goto pasv_error;
     }
-  seteuid ((uid_t) cred.uid);
+  if (seteuid ((uid_t) cred.uid))
+    _exit (EXIT_FAILURE);
   pasv_addrlen = sizeof (pasv_addr);
   if (getsockname (pdata, (struct sockaddr *) &pasv_addr, &pasv_addrlen) < 0)
     goto pasv_error;
   if (listen (pdata, 1) < 0)
     goto pasv_error;
 
-  if (pasv_addr.ss_family == AF_INET6)
+  if (epsv == PASSIVE_EPSV)
     {
-      reply (229, "Extended Passive Mode OK (|||%u|)",
-	     ((struct sockaddr_in6 *) &pasv_addr)->sin6_port);
+      /* EPSV for IPv4 and IPv6.  */
+      reply (229, "Entering Extended Passive Mode (|||%u|)",
+	     ntohs((pasv_addr.ss_family == AF_INET)
+		    ? ((struct sockaddr_in *) &pasv_addr)->sin_port
+		    : ((struct sockaddr_in6 *) &pasv_addr)->sin6_port));
       return;
     }
-  else
+  else /* !EPSV */
     {
-      a = (char *) &((struct sockaddr_in *) &pasv_addr)->sin_addr;
-      p = (char *) &((struct sockaddr_in *) &pasv_addr)->sin_port;
+      /* PASV for IPv4, or LPSV for IPv4 or IPv6.
+       *
+       * Some systems, like OpenSolaris, prefer to return
+       * an IPv4-mapped-IPv6 address, which must be processed
+       * for printout.  */
 
 #define UC(b) (((int) b) & 0xff)
 
-      reply (227, "Entering Passive Mode (%d,%d,%d,%d,%d,%d)", UC (a[0]),
-	     UC (a[1]), UC (a[2]), UC (a[3]), UC (p[0]), UC (p[1]));
+      if (pasv_addr.ss_family == AF_INET6
+	  && IN6_IS_ADDR_V4MAPPED (&((struct sockaddr_in6 *) &pasv_addr)->sin6_addr))
+	{
+	  a = (char *) &((struct sockaddr_in6 *) &pasv_addr)->sin6_addr;
+	  a += 3 * sizeof (struct in_addr);	/* Skip padding up to IPv4 content.  */
+	  p = (char *) &((struct sockaddr_in6 *) &pasv_addr)->sin6_port;
+	}
+      else if (pasv_addr.ss_family == AF_INET6)
+	{
+	  /* LPSV for IPv6, not mapped. */
+	  a = (char *) &((struct sockaddr_in6 *) &pasv_addr)->sin6_addr;
+	  p = (char *) &((struct sockaddr_in6 *) &pasv_addr)->sin6_port;
+
+	  reply (228, "Entering Long Passive Mode "
+		 "(6,16,%d,%d,%d,%d,%d,%d,%d,%d"	/* a[0..7] */
+		 ",%d,%d,%d,%d,%d,%d,%d,%d"	/* a[8..15] */
+		 ",2,%d,%d)",	/* p0, p1 */
+		 UC (a[0]), UC (a[1]), UC (a[2]), UC (a[3]),
+		 UC (a[4]), UC (a[5]), UC (a[6]), UC (a[7]),
+		 UC (a[8]), UC (a[9]), UC (a[10]), UC (a[11]),
+		 UC (a[12]), UC (a[13]), UC (a[14]), UC (a[15]),
+		 UC (p[0]), UC (p[1]));
+	  return;
+	}
+      else
+	{
+	  a = (char *) &((struct sockaddr_in *) &pasv_addr)->sin_addr;
+	  p = (char *) &((struct sockaddr_in *) &pasv_addr)->sin_port;
+	}
+
+      if (epsv == PASSIVE_LPSV)
+	reply (228, "Entering Long Passive Mode "
+	       "(4,4,%d,%d,%d,%d,2,%d,%d)",
+	       UC (a[0]), UC (a[1]), UC (a[2]), UC (a[3]),
+	       UC (p[0]), UC (p[1]));
+      else
+	reply (227, "Entering Passive Mode (%d,%d,%d,%d,%d,%d)",
+	       UC (a[0]), UC (a[1]), UC (a[2]), UC (a[3]),
+	       UC (p[0]), UC (p[1]));
       return;
     }
 
