@@ -1,7 +1,7 @@
 /*
   Copyright (C) 1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003,
-  2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012 Free Software
-  Foundation, Inc.
+  2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014,
+  2015 Free Software Foundation, Inc.
 
   This file is part of GNU Inetutils.
 
@@ -55,13 +55,7 @@
 #include <time.h>
 #include <netinet/in.h>
 #include <sys/select.h>
-#ifdef HAVE_OSOCKADDR_H
-# include <osockaddr.h>
-#else
-# ifndef HAVE_STRUCT_OSOCKADDR
-#  include <protocols/osockaddr.h>
-# endif
-#endif
+#include <protocols/osockaddr.h>
 #include <protocols/talkd.h>
 #include <errno.h>
 #include "talk_ctl.h"
@@ -82,8 +76,12 @@ ctl_transact (struct in_addr target, CTL_MSG msg, int type, CTL_RESPONSE * rp)
   struct timeval wait;
 
   msg.type = type;
+  daemon_addr.sin_family = AF_INET;
+#ifdef HAVE_STRUCT_SOCKADDR_IN_SIN_LEN
+  daemon_addr.sin_len = sizeof (daemon_addr);
+#endif
   daemon_addr.sin_addr = target;
-  daemon_addr.sin_port = daemon_port;
+  daemon_addr.sin_port = htons (daemon_port);
   FD_ZERO (&ctl_mask);
   FD_SET (ctl_sockt, &ctl_mask);
 
@@ -93,11 +91,11 @@ ctl_transact (struct in_addr target, CTL_MSG msg, int type, CTL_RESPONSE * rp)
    */
   do
     {
-      wait.tv_sec = CTL_WAIT;
-      wait.tv_usec = 0;
       /* resend message until a response is obtained */
       do
 	{
+	  wait.tv_sec = CTL_WAIT;
+	  wait.tv_usec = 0;
 	  cc = sendto (ctl_sockt, (char *) &msg, sizeof (msg), 0,
 		       (struct sockaddr *) &daemon_addr,
 		       sizeof (daemon_addr));
